from django_object_authority.utils import get_full_permission_codename


class BaseFilter(object):
    """A base class from which all filter backend classes should inherit."""

    def filter_queryset(self, request, queryset, view):
        """Return a filtered queryset."""
        raise NotImplementedError(".filter_queryset() must be overridden.")


class AuthorityBaseFilter(BaseFilter):
    """
    Project filter backend restrict items on queryset according the user permissions.
    The permissions of who request it are provided for his groups. Note that is required define three specific
    permissions for the model.
    Note that is important check his permission with an specific order (defined on: `permission_codes`).
    """
    permission_codes = ()
    prefix_filter_method = 'filter_by'

    def filter_queryset(self, request, queryset, view):
        """Iterate over `permission_codes` to check and filter according its permission codes."""

        if not self._is_valid():
            raise NotImplementedError("Some of {} methods are not implemented.".format(
                self._get_requied_filter_methods()))

        model = getattr(view, 'model', None)
        for code in self.permission_codes:
            # check permission by codename
            codename = get_full_permission_codename(code, getattr(model, '_meta', None))
            if request.user.has_perm(codename, None):
                # filter by criteria
                filter_method = self._get_filter_method(code)
                if hasattr(self, filter_method):
                    queryset = getattr(self, filter_method)(queryset, request.user)

        return queryset

    def _get_filter_method(self, code):
        """Format the filter method according the current permission code."""
        return '{}_{}'.format(self.prefix_filter_method, code)

    def _get_requied_filter_methods(self):
        """Return all required methods that should be implemented."""
        return [self._get_filter_method(code) for code in self.permission_codes]

    def _is_valid(self):
        """Validating that required methods are implemented."""
        filter_methods = self._get_requied_filter_methods()
        return filter_methods and all(map(lambda method: hasattr(self, method), filter_methods))


from . import settings
from .utils import get_full_permission_codename


class BaseObjectAuthorization(object):

    def has_object_permission(self, user, obj):
        return settings.AUTHORIZE_BY_DEFAULT


class BaseUserObjectAuthorization(BaseObjectAuthorization):

    def has_add_permission(self, user, obj):
        return self._has_permission(user, 'add', obj)

    def has_change_permission(self, user, obj):
        return self._has_permission(user, 'change', obj)

    def has_delete_permission(self, user, obj):
        return self._has_permission(user, 'delete', obj)

    def has_view_permission(self, user, obj):
        return self._has_permission(user, 'view', obj)

    def _has_permission(self, user, action, obj):
        permission_codename = get_full_permission_codename(action, obj._meta)
        return user.has_perm(permission_codename)



from . import settings


class AuthorizationMixin(object):
    """
    Mixin for filter queryset according the defined list of backend.

    :keyword filter_backends: filter backend list
    """

    authorization_filter_class = None

    def get_authorization_filter_class(self):
        """Return all available backends"""
        return self.authorization_filter_class

    def get_queryset(self):
        """
        Method that tries to get a filtered queryset from their parent if it has method defined.
        Otherwise get the queryset from de default manager of the model or raise an error.
        """
        try:
            queryset = super().get_queryset()
        except AttributeError:
            from django.core.exceptions import ImproperlyConfigured
            if self.model is not None:
                queryset = self.model._default_manager.all()
            else:
                raise ImproperlyConfigured(
                    "{cls} is missing a QuerySet. Define {cls}.model, {cls}.queryset, or override "
                    "{cls}.get_queryset().".format({'cls': self.__class__.__name__})
                )
        # no filter queryset if the user is superuser or staff and has active the correct setting
        if self.full_permission_for_superuser() or self.full_permission_for_staff():
            return queryset
        return self.filter_queryset(queryset)

    def full_permission_for_superuser(self):
        """Return if the user is superuser and has full permission activated."""
        return self.request.user.is_superuser and settings.FULL_PERMISSION_FOR_SUPERUSERS

    def full_permission_for_staff(self):
        """Return if the user is staff and has full permission activated."""
        return self.request.user.is_staff and settings.FULL_PERMISSION_FOR_STAFF

    def filter_queryset(self, queryset):
        """Given a queryset, filter it with whichever filter backend is in use."""
        filter_class = self.get_authorization_filter_class()
        if filter_class:
            queryset = filter_class().filter_queryset(self.request, queryset, self)
        return queryset




import copy
from importlib import import_module


def autodiscover_authorizations(*args, **kwargs):
    """
    Auto-discover authorization modules and fail silently when not present.
    This forces an import on them to register any authorization classes.

    You may provide a register_to keyword parameter as a way to access a
    registry. This register_to object must have a _registry instance variable
    to access it.
    """
    from django.apps import apps

    register_to = kwargs.get('register_to')
    for app_config in apps.get_app_configs():
        for module_to_search in args:
            # Attempt to import the authorization's module.
            if register_to:
                before_import_registry = copy.copy(register_to._registry)
            try:
                import_module('{}.{}'.format(app_config.name, module_to_search))
            except Exception:
                # Restore last registry on import error occurred
                if register_to:
                    register_to._registry = before_import_registry



from django.db.models.base import ModelBase


class AlreadyAuthorizationRegistered(Exception):
    pass


class NoAuthorizationRegistered(Exception):
    pass


class Authorization(object):
    """
    A Authorization object encapsulates an instance of a object permission.
    Models are registered with the Authorization using the register() method.
    Use unregister() method to clean Authorization classes registered.
    """
    def __init__(self):
        self._registry = {}

    def register(self, model_or_iterable, authorization_class=None, **options):
        """Register model authorization"""
        if isinstance(model_or_iterable, ModelBase):
            model_or_iterable = [model_or_iterable]
        for model in model_or_iterable:
            if model in self._registry and self._registry[model].registered:
                raise AlreadyAuthorizationRegistered(
                    'Model "{}" is already registered for authorization'.format(model.__name__)
                )
            try:
                self._register(model, authorization_class)
            except Exception:
                self._registry[model].registered = False
                raise

    def unregister(self, model_or_iterable):
        """Unregister model authorization"""
        if isinstance(model_or_iterable, ModelBase):
            model_or_iterable = [model_or_iterable]
        for model in model_or_iterable:
            self._registry.pop(model)

    def get(self, model):
        """Get registered model"""
        try:
            return self._registry[model]
        except KeyError:
            raise NoAuthorizationRegistered(
                'Model "{}" is not registered for authorization'.format(model.__name__)
            )

    def _register(self, model, authorization_class):
        setattr(authorization_class, 'registered', True)
        self._registry[model] = authorization_class


# This global object represents the default authorization storage registry.
authorization = Authorization()



from django.apps import apps
from django.contrib.auth import get_permission_codename

from django_object_authority.settings import DEFAULT_PERMISSIONS, PERMISSION_FOR_APPLICATIONS, PERMISSION_FOR_MODELS


def create_update_permissions(applications=None, models=None, permissions=None, **kwargs):
    """
    Create the default permissions (view, add, change, delete) for each model of application or those models
    you specify in settings or as local parameter of those application also specified in settings or as local parameter.

    :param applications: list of application names
    :param models: list of model names
    :param permissions: list of permission you want to create or update
    :param kwargs: extra args that post-migrate signal sends.
    :return:
    """

    created_perm, updated_perm = 0, 0

    # get default applications from settings or all from installed apps.
    default_apps = PERMISSION_FOR_APPLICATIONS or apps.all_models.keys()
    available_applications = set(default_apps)
    if applications:
        available_applications = set(applications) & set(apps.all_models.keys())

    # get filtering models
    default_models = PERMISSION_FOR_MODELS or models
    available_models = set()
    for app in available_applications:
        # iterate for all available applications to filter by models.
        # filtering expected a list of models and those models don't belong to many to many model relations.
        app_models = set()
        for model_name, model in apps.all_models.get(app).items():
            included = default_models is None or (default_models and model_name in set(default_models))
            if included and not is_autogenerated(model):
                app_models.add(model)

        # set all application models if filter is not defined or not apply
        if not app_models:
            app_models = set(apps.all_models.get(app).values())
        available_models.update(app_models)

    # Update default permissions with the list specified as parameter
    permissions = DEFAULT_PERMISSIONS if not permissions else list(permissions, ) + list(DEFAULT_PERMISSIONS)

    for model in available_models:
        result = _create_permissions(model, permissions)
        created_perm += result[0]
        updated_perm += result[1]

    return created_perm, updated_perm


def _create_permissions(model, permissions):
    """Get permission or create it."""
    from django.contrib.auth.models import Permission
    from django.contrib.contenttypes.models import ContentType

    created, updated = 0, 0
    if permissions is not None:
        for perm in permissions:
            kwargs = {
                'codename': "{}_{}".format(perm, model._meta.model_name),
                'content_type': ContentType.objects.get_for_model(model)
            }
            defaults = {'name': "Can {} {}".format(perm.replace('_', ' '), model._meta.verbose_name)}
            obj, _created = Permission.objects.get_or_create(defaults=defaults, **kwargs)

            # count number of permission has created
            if _created:
                created += 1
            else:
                updated += 1

    return created, updated


def get_full_permission_codename(action, opts):
    """
    Returns the full codename of the permission for the specified action.
    """
    app_label = getattr(opts, 'app_label', '')
    return '{}.{}'.format(app_label, get_permission_codename(action, opts))


def is_autogenerated(model):
    """
    Check if model is auto generated.
    Is supposed that auto generated django models always has underscore in his name to separate two linked tables and
    there is not possible define a custom model with underscores.

    :param model:
    :return:
    """
    return '_' in model._meta.model_name


from django.core.management.base import BaseCommand

from django_object_authority.utils import create_update_permissions


APPS = 'applications'
MODELS = 'models'
PERMISSIONS = 'permissions'


class Command(BaseCommand):
    help = "Utility to create or update all required permissions to manage the app."

    def add_arguments(self, parser):
        parser.add_argument('-a', '--applications', nargs='+', type=str, dest=APPS, default=None,
                            help="List of applications to create permissions of all their models.")
        parser.add_argument('-m', '--models', nargs='+', type=str, dest=MODELS, default=None,
                            help="List of models to create permissions.")
        parser.add_argument('-n', '--new', nargs='+', type=str, dest=PERMISSIONS, default=None,
                            help="List of new permission labels.")

    def handle(self, *args, **options):
        self.initialize(options)
        self.create_update_permissions()
        self.stdout.write(self.style.SUCCESS("Permission creation has successfully finished!"))

    def initialize(self, options):
        """Set all options as instance attribute."""
        for key, value in options.items():
            setattr(self, key, value)

    def create_update_permissions(self):
        """Create permissions for all available application models."""
        self.stdout.write("Creating permissions...")
        created_perm, updated_perm = create_update_permissions(applications=getattr(self, APPS),
                                                               models=getattr(self, MODELS),
                                                               permissions=getattr(self, PERMISSIONS))
        self.stdout.write("There are {} new permission and ignore {} because they already exist!".format(
            created_perm, updated_perm))




def register(*models, **kwargs):
    """
    Registers the given model(s) classes into authorization object:

    @authorizations.register(Project)
    class ProjectAuthorization(options.BaseObjectAuthorization):
        pass
    """
    from .authorizations import Authorization, authorization

    def _authorization_wrapper(authorization_class):
        if not models:
            raise ValueError('At least one model must be passed to register.')
        authorization_obj = kwargs.pop('authorization', authorization)
        if not isinstance(authorization_obj, Authorization):
            raise ValueError('site must subclass Authorization')
        authorization_obj.register(models, authorization_class=authorization_class)
        return authorization_class
    return _authorization_wrapper



from django.apps import AppConfig
from django.db.models.signals import post_migrate
from django.utils.translation import ugettext_lazy as _

from .utils import create_update_permissions


class DjangoObjectAuthorityConfig(AppConfig):
    name = 'django_object_authority'
    verbose_name = _("Django object authority")

    def ready(self):
        """Register all available authorization from modules."""
        super().ready()
        self.module.autodiscover()
        # connect to the event for migration command
        post_migrate.connect(create_update_permissions, sender=self)



from .authorizations import authorization, NoAuthorizationRegistered, AlreadyAuthorizationRegistered
from .decorators import register
from .filters import AuthorityBaseFilter, BaseFilter
from .loader import autodiscover_authorizations
from .mixins import AuthorizationMixin
from .options import BaseObjectAuthorization, BaseUserObjectAuthorization


__version__ = "0.0.5"


default_app_config = 'django_object_authority.apps.DjangoObjectAuthorityConfig'


__all__ = [
    'AuthorityBaseFilter', 'AlreadyAuthorizationRegistered', 'AuthorizationMixin', 'BaseFilter',
    'BaseObjectAuthorization', 'BaseUserObjectAuthorization', 'NoAuthorizationRegistered', 'authorization',
    'autodiscover', 'register',
]


def autodiscover():
    autodiscover_authorizations('authorizations', register_to=authorization)



"""
Settings for Django object authority are all namespaced in the OBJECT_AUTHORITY setting.
For example your project's `settings.py` file might look like this:

OBJECT_AUTHORITY_AUTHORIZE_BY_DEFAULT = True
OBJECT_AUTHORITY_FULL_PERMISSION_FOR_STAFF = True
OBJECT_AUTHORITY_FULL_PERMISSION_FOR_SUPERUSERS = True
"""

from django.conf import settings

AUTHORIZE_BY_DEFAULT = getattr(settings, 'OBJECT_AUTHORITY_AUTHORIZE_BY_DEFAULT', True)
FULL_PERMISSION_FOR_STAFF = getattr(settings, 'OBJECT_AUTHORITY_FULL_PERMISSION_FOR_STAFF', False)
FULL_PERMISSION_FOR_SUPERUSERS = getattr(settings, 'OBJECT_AUTHORITY_FULL_PERMISSION_FOR_SUPERUSERS', True)
CHECK_PERMISSION_CLASS_BY_DEFAULT = getattr(settings, 'OBJECT_AUTHORITY_CHECK_PERMISSION_CLASS_BY_DEFAULT', True)

# Settings applied on post migrations to create permission only for specified apps and models
# or for all models of all installed apps.

PERMISSION_FOR_MODELS = getattr(settings, 'OBJECT_AUTHORITY_PERMISSION_FOR_MODELS', None)
PERMISSION_FOR_APPLICATIONS = getattr(settings, 'OBJECT_AUTHORITY_PERMISSION_FOR_APPLICATIONS', None)


# Default permissions that will be create for models specified before.
DEFAULT_PERMISSIONS = ('view', 'add', 'change', 'delete')



import re

from django.contrib.auth.backends import ModelBackend
from django.utils.text import slugify

from . import settings, authorization, NoAuthorizationRegistered, BaseObjectAuthorization


class ObjectAuthorityBackend(ModelBackend):

    def has_perm(self, user_obj, perm, obj=None):
        """Main method wrapper to check permissions."""
        return self.has_object_perm(user_obj, perm, obj)

    def has_object_perm(self, user_obj, perm=None, obj=None):
        """
        Validate permissions for the object specified. If there is not object should be validated before with
        the `contrib.auth.backends.ModelBackend`.

        Is required class permission to has permission over an object of the class.
        permissions: `change` `delete` `view`
        """
        if not user_obj.is_active or user_obj.is_anonymous or not all([perm, obj]):
            return False

        # validation object permission
        obj_permission_label = '_has_perm_cache_{}_{}'.format(slugify(perm), obj.id)
        if not hasattr(user_obj, obj_permission_label):
            setattr(user_obj, obj_permission_label, self._has_object_permissions(user_obj, perm, obj))
        return getattr(user_obj, obj_permission_label)

    def _has_object_permissions(self, user_obj, perm, obj):
        """Gets the user permission for the object.

        By default, allow all staff and superusers do the action:
        Settings configuration:
        * FULL_PERMISSION_FOR_STAFF
        * FULL_PERMISSION_FOR_SUPERUSERS

        Gets the registered permission class linked with the object class and validate for the action specified.
        """
        model = self._get_model(obj)
        assert model is not None, (
            "Object {} must be a `Model` instance to check object permission.".format_map(obj.__class__)
        )

        # check if is superuser or staff to allow the access to the object
        if self._has_permission_as_superuser(user_obj) or self._has_permission_as_staff(user_obj):
            return True

        action = self._get_action(perm, model)
        model_permission = self._get_model_permission_class(model)

        if action:
            action_method_name = "has_{action}_permission".format(action=action)
            if hasattr(model_permission, action_method_name):
                return getattr(model_permission, action_method_name)(user=user_obj, obj=obj)

        # check class permission if CHECK_PERMISSION_CLASS_BY_DEFAULT is True
        if settings.CHECK_PERMISSION_CLASS_BY_DEFAULT:
            return user_obj.has_perm(perm)

        # otherwise check default object permission
        if hasattr(model_permission, 'has_object_permission'):
            return model_permission.has_object_permission(user_obj, obj)

        return False

    def _get_model_permission_class(self, model):
        """
        Gets the permission class registered for the current model.

        Return default `BaseObjectPermission` class if there are no permission classes regitered for the model.
        """
        try:
            return authorization.get(model)()
        except NoAuthorizationRegistered:
            return BaseObjectAuthorization()

    def _get_action(self, perm, model):
        try:
            perm = perm.rsplit('.', 1)[1]
        except IndexError:
            pass
        try:
            return re.sub(r'_{}.*'.format(self._get_model_name(model)), '', perm)
        except (TypeError, IndexError):
            return None

    @staticmethod
    def _get_model(obj):
        return getattr(getattr(obj, '_meta', object()), 'model', None)

    @staticmethod
    def _get_model_name(obj):
        return getattr(getattr(obj, '_meta', object()), 'model_name', '')

    @staticmethod
    def _get_model_label(obj):
        return getattr(getattr(obj, '_meta', object()), 'label', '')

    @staticmethod
    def _has_permission_as_staff(user):
        return getattr(user, 'is_staff', False) and settings.FULL_PERMISSION_FOR_STAFF

    @staticmethod
    def _has_permission_as_superuser(user):
        return getattr(user, 'is_superuser', False) and settings.FULL_PERMISSION_FOR_SUPERUSERS


class DRFAuthorityBackend(ObjectAuthorityBackend):
    """
    A base class permission backend with required methods from `BasePermission` drf class.
    """
    default_action_mapping = {

        # default mapping for APIView subclasses
        'get': 'view',
        'post': 'add',
        'put': 'change',
        'patch': 'change',
        'delete': 'delete',

        # default mapping for GenericViewSet subclasses
        'retrieve': 'view',
        'create': 'add',
        'update': 'change',
        'partial_update': 'change',
        'destroy': 'delete',
    }

    def has_permission(self, request, view):
        return self.has_object_permission(request, view, None)

    def has_object_permission(self, request, view, obj):
        action = getattr(view, 'action', request.method.lower())
        if action not in self.default_action_mapping:
            return False
        return self.has_perm(request.user, self.default_action_mapping[action], obj)
